/* THIS FILE IS AUTOGENERATED - DO NOT EDIT */

#ifndef mozilla_dom_RequestBinding_h
#define mozilla_dom_RequestBinding_h

#include "js/RootingAPI.h"
#include "jspubtd.h"
#include "mozilla/ErrorResult.h"
#include "mozilla/dom/BindingDeclarations.h"
#include "mozilla/dom/Nullable.h"
#include "mozilla/dom/UnionTypes.h"

namespace mozilla {
namespace dom {

class ArrayBufferOrArrayBufferViewOrBlobOrScalarValueStringOrURLSearchParams;
class HeadersOrByteStringSequenceSequenceOrByteStringMozMap;
struct NativePropertyHooks;
class OwningArrayBufferOrArrayBufferViewOrBlobOrScalarValueStringOrURLSearchParams;
class OwningHeadersOrByteStringSequenceSequenceOrByteStringMozMap;
class ProtoAndIfaceCache;
class Request;
struct RequestAtoms;
struct RequestInitAtoms;

} // namespace dom
} // namespace mozilla

namespace mozilla {
namespace dom {

MOZ_BEGIN_ENUM_CLASS(RequestMode, uint32_t)
  Same_origin,
  No_cors,
  Cors,
  EndGuard_
MOZ_END_ENUM_CLASS(RequestMode)

namespace RequestModeValues {
extern const EnumEntry strings[4];
} // namespace RequestModeValues


MOZ_BEGIN_ENUM_CLASS(RequestCredentials, uint32_t)
  Omit,
  Same_origin,
  Include,
  EndGuard_
MOZ_END_ENUM_CLASS(RequestCredentials)

namespace RequestCredentialsValues {
extern const EnumEntry strings[4];
} // namespace RequestCredentialsValues


struct RequestInit : public DictionaryBase
{
  Optional<OwningArrayBufferOrArrayBufferViewOrBlobOrScalarValueStringOrURLSearchParams> mBody;
  Optional<RequestCredentials> mCredentials;
  Optional<OwningHeadersOrByteStringSequenceSequenceOrByteStringMozMap> mHeaders;
  Optional<nsCString> mMethod;
  Optional<RequestMode> mMode;

  RequestInit();

protected:
  explicit inline RequestInit(int )
  {
    // Do nothing here; this is used by our "Fast" subclass
  }

private:
  RequestInit(const RequestInit&) MOZ_DELETE;
  void operator=(const RequestInit) MOZ_DELETE;

  static bool
  InitIds(JSContext* cx, RequestInitAtoms* atomsCache);

public:
  bool
  Init(JSContext* cx, JS::Handle<JS::Value> val, const char* sourceDescription = "Value", bool passedToJSImpl = false);

  bool
  Init(const nsAString& aJSON);

  bool
  ToObjectInternal(JSContext* cx, JS::MutableHandle<JS::Value> rval) const;

  bool
  ToJSON(nsAString& aJSON) const;

  void
  TraceDictionary(JSTracer* trc);
};

namespace binding_detail {
struct FastRequestInit : public RequestInit
{
  inline FastRequestInit()
    : RequestInit(42)
  {
    // Doesn't matter what int we pass to the parent constructor
  }
};
} // namespace binding_detail


namespace RequestBinding {

  typedef mozilla::dom::Request NativeType;

  // We declare this as an array so that retrieving a pointer to this
  // binding's property hooks only requires compile/link-time resolvable
  // address arithmetic.  Declaring it as a pointer instead would require
  // doing a run-time load to fetch a pointer to this binding's property
  // hooks.  And then structures which embedded a pointer to this structure
  // would require a run-time load for proper initialization, which would
  // then induce static constructors.  Lots of static constructors.
  extern const NativePropertyHooks sNativePropertyHooks[];

  JSObject*
  DefineDOMInterface(JSContext* aCx, JS::Handle<JSObject*> aGlobal, JS::Handle<jsid> id, bool aDefineOnGlobal);

  bool
  ConstructorEnabled(JSContext* aCx, JS::Handle<JSObject*> aObj);

  const JSClass*
  GetJSClass();

  JSObject*
  Wrap(JSContext* aCx, mozilla::dom::Request* aObject, nsWrapperCache* aCache);

  template <class T>
  inline JSObject* Wrap(JSContext* aCx, T* aObject)
  {
    return Wrap(aCx, aObject, aObject);
  }

  void
  CreateInterfaceObjects(JSContext* aCx, JS::Handle<JSObject*> aGlobal, ProtoAndIfaceCache& aProtoAndIfaceCache, bool aDefineOnGlobal);

  JS::Handle<JSObject*>
  GetProtoObjectHandle(JSContext* aCx, JS::Handle<JSObject*> aGlobal);

  JS::Handle<JSObject*>
  GetConstructorObjectHandle(JSContext* aCx, JS::Handle<JSObject*> aGlobal, bool aDefineOnGlobal = true);

  JSObject*
  GetConstructorObject(JSContext* aCx, JS::Handle<JSObject*> aGlobal);

} // namespace RequestBinding



} // namespace dom
} // namespace mozilla

#endif // mozilla_dom_RequestBinding_h
